#include <u.h>
#include <libc.h>
#include <bio.h>
#include <draw.h>
#include <html.h>
#include <ctype.h>
#include "dat.h"

enum {
	WideTableField = 60,	/* table fields wider than this get wrapped */
	Verbatim = 0,			/* layout preserved, no line wrap, no justification */
};

typedef struct Xlate Xlate;
struct Xlate {
	Fmt fmt;
	Fmt *f;				/* pointer to the above - just makes the syntax cleaner */
	int width;			/* width of this block (may be a table cell) */
	int indent;			/* left indent in 10ths of a tabstop */
	int just;			/* justification mode l/r/c/0 0==Verbatim */
	int font;			/* font ID */
	int pos;			/* output position (column) */
	int embed;			/* true if we prever formatting embedded */
	int curanchor;		/* current anchor ID */
	URLwin *u;			/* urlwin passed in */
	int ul;				/* underline mdoe */
	int vshift;			/* vert shift in pica */
};

static void walk(Xlate *x, Item *items);


/* at start of line */
static int
sol(Xlate *x)
{
	return x->pos == 0;
}


/* escape some chars that have special meaning at the start of a line */
static void
esc(Xlate *x, Rune c)
{
	if(c == '\\')
		fmtprint(x->f, "\(bs");
	else
	if(x->pos == 0 && c == L'.')
		fmtprint(x->f, "\\&.");
	else
		fmtprint(x->f, "%C", c);
}

/*
 * used for prose text which can be re-flowed
 *
 * mutate all whitespace to space characters,
 * drop whitespace sta start of lines,
 * wrap lines on a space char
 */
static void
wrap(Xlate *x, Rune *r)
{
	Rune c;

	if(debug)
		fprint(2, "wrap %3d '%S'\n", x->pos, r);

	for( ;*r; r++){
		c = *r;

		if(c == L'­'){							/* soft (optional) hyphen */
			fmtprint(x->f, "\\%%");
			x->pos+= 2;
			continue;
		}

		if(isspacerune(c)){
			c = ' ';
			if(x->pos == 0)
				continue;
			if(x->pos > x->width){				/* width from command line */
				fmtprint(x->f, "\n");
				x->pos = 0;
				continue;
			}
		}

		esc(x, c);
		x->pos++;
	}
}

/* used to output verbatim text */
static void
copy(Xlate *x, Rune *r)
{
	Rune c;

	if(debug)
		fprint(2, "copy %3d '%S'\n", x->pos, r);

	for( ;*r; r++){
		c = *r;

		if(c == L'\n')
			x->pos = 0;

		esc(x, c);
		x->pos++;
	}
}

/* write raw data (no escaping) */
static void
raw(Xlate *x, char *str)
{
	char *p;

	for(p = str ;*p; p++){
		switch(*p){
		case '\n':
			x->pos = 0;
			break;
		default:
			x->pos++;
			break;
		}
		fmtprint(x->f, "%c", *p);
	}
}

/* linebreak (conditional newline) */
static void
lnbrk(Xlate *x)
{
	if(! sol(x)){
		fmtprint(x->f, "\n");
		x->pos = 0;
	}
}

/*
 * change font face or/and size 
 */
static void
setfont(Xlate * x, int new)
{
    int old;
    char *embed, *colzero;

    old = x->font;
    if(old / NumSize != new / NumSize){
		switch(new / NumSize){
		case FntI:
		    embed = "\\fI";
		    colzero = ".I\n";
		    break;
		case FntB:
		    embed = "\\fB";
		    colzero = ".B\n";
		    break;
		case FntT:
		    embed = "\\f(CW";
		    colzero = ".CW\n";
		    break;
		default:		/* FATHRU */
		case FntR:
		    embed = "\\fR";
		    colzero = ".R\n";
		    break;
		}
		if(sol(x) && !x->embed)
		    raw(x, colzero);
		else
		    raw(x, embed);
    }

    if (old % NumSize != new % NumSize){
		switch(new % NumSize){
		case Tiny:
		    embed = "\\s6";
		    colzero = ".ps 6\n";
		    break;
		case Small:
		    embed = "\\s8";
		    colzero = ".ps 8\n";
		    break;
		case Large:
		    embed = "\\s(14";
		    colzero = ".ps 14\n";
		    break;
		case Verylarge:
		    embed = "\\s(18";
		    colzero = ".ps 18\n";
		    break;
		default:		/* FATHRU */
		case Normal:
		    embed = "\\s(10";
		    colzero = ".ps 10\n";
		    break;
		}
		if(sol(x) && !x->embed)
		    raw(x, colzero);
		else

		    raw(x, embed);
    }
	x->font = new;
}

/* write text blocks */
static void
text(Xlate *x, Item *item)
{
	Itext *it;
	char buf[64];
	int state, just;

	state = item->state;
	just = state & (IFcjust|IFrjust|IFwrap);
	it = (Itext *)item;

	/* justrification/centering/display */
	if(just != x->just && !x->embed){
		lnbrk(x);						/* somthing in fill/justify has changed, stop previous mode */
		if(x->just & IFcjust){
 			raw(x, ".ce 0\n");
 			raw(x, ".LP\n");
		}

		if(x->just == Verbatim){
			raw(x, ".NL\n");
			raw(x, ".DE\n");
		}

		if(state & IFcjust){				/* start centered text */
			raw(x, ".ce 99\n");
		}
		else
		if(state == Verbatim){			/* start verbatim text */
			raw(x, ".DS\n");
			raw(x, ".ps -2\n");
		}
		else{							/* start left justified and filled */
			/* nothing special to do */
		}
		x->just = just;
	}

	if(state & IFhang){
		x->indent = (state & IFhangmask) + (((state & IFhangmask) >> IFindentshift) *10);
		snprint(buf, sizeof(buf), "\n\n[%d]\n\n", x->indent);
		raw(x, buf);
	}

	/* switch font face or/and size */
	if(!x->embed)
		setfont(x, it->fnt);


	/* underline and strikethru */
	if(it->ul != x->ul){
		x->ul = it->ul;
		switch(x->ul){
		case ULunder:
			lnbrk(x);
			raw(x, ".cu 100\n");
		break;
		case ULnone:
			lnbrk(x);
			raw(x, ".cu 0\n");
			break;
		default:
			break;
		}
	}

	/* subpixel vertical adjustment */
	if((it->voff -Voffbias)/2 != 0){
		snprint(buf, sizeof(buf), "\\v'%dp'", ((int)it->voff -Voffbias)/2);
		raw(x, buf);
	}

	if(state & IFbrksp){				/* insert a blank line */
		lnbrk(x);
		if(x->just == Verbatim)			/* in verbation newlines are newlines */
			raw(x, "\n");
		else
		if(x->just & IFcjust)			/* cannot use .LP in a centered block */
			raw(x, ".sp 1\n");
		else
		if(state & IFhang){
			snprint(buf, sizeof(buf), ".IP \\(bu\n");
			raw(x, buf);
			return;
		}
		else
			raw(x, ".LP\n");			/* must be normal text */
	}
	else
	if(state & IFbrk){					/* break current line */
		if(! sol(x))
			if(just == Verbatim)		/* in verbation just a newline */
				raw(x, "\n");
			else{
				lnbrk(x);
				raw(x, ".br\n");
			}
	}

	if(just == Verbatim)
		copy(x, it->s);
	else
		wrap(x, it->s);

	/* restore subpixel vertical adjustment */
	if((it->voff -Voffbias)/2 != 0){
		snprint(buf, sizeof(buf), "\\v'-%dp'", ((int)it->voff -Voffbias)/2);
		raw(x, buf);
	}
}

/* write a table cell */
static int
cell(Xlate *x, Item *item, int width, int idx, int nowrap)
{
	char *str;
	Xlate xlate;

	xlate = *x;
	xlate.f = &xlate.fmt;
	xlate.width = width;
	xlate.embed = 1;
	xlate.pos = 0;

	fmtstrinit(xlate.f);
	walk(&xlate, item);
	str = fmtstrflush(xlate.f);

	if(strlen(str) > WideTableField && !nowrap){
		if(idx == 0){		/* first line */
			raw(x, "T{\n");
			raw(x, str);
			raw(x, "\nT}");
		}
		else{
			raw(x, "\tT{\n");
			raw(x, str);
			raw(x, "\nT}");
		}
	}
	else{
		if(idx == 0){		/* first line */
			raw(x, str);
		}
		else{
			raw(x, "\t");
			raw(x, str);
		}
	}

	free(str);
	return 1;
}

static int
dimen(Xlate *x, Dimen d)
{
	int pt;

	pt = -1;
	switch(dimenkind(d)){
	case Dpixels:
		pt = dimenspec(d) * 8;					/* FIXME: 8 == absolute guesswork */
		break;
	case Dpercent:
		pt = (dimenspec(d) *x->width *6)/100;	/* FIXME: 6 seems all wrong, expected 72 */
		break;
	case Drelative:
		pt = -1;								/* not done this yet */
		break;
	case Dnone:
		pt = -1;
		break;
	default:
		fprint(2, "unknown dimension %d,%d\n", dimenkind(d), dimenspec(d));
		break;
	}
	return pt;
}


static void
rowdef(Xlate *x, Tablecell *ce)
{
	int i, pt;
	char *just, buf[16];

	switch(ce->align.halign){
	case ALleft:
		just = "l";
		break;
	case ALright:
		just = "r";
		break;
	case ALcenter:
		just = "c";
		break;
	default:
		fprint(2, "%s: unknown table alignment %x\n", argv0, ce->align.halign);
		just = "l";
		break;
	}
		
	if((pt = dimen(x, ce->wspec)) != -1)
		snprint(buf, sizeof(buf), "%sW(%dp)", just, pt);
	else
		snprint(buf, sizeof(buf), "%s", just);
	raw(x, buf);

	/* column span */
	for(i = 1; i < ce->rowspan; i++)
		raw(x, "|s");
}

/*
 * are the remaining table rows the same
 *
 * NB: remember the number of row cells may be less than tab->nrow
 * due to cells spanning multiple colums
 */
int
allsame(Table *tab, int r1)
{
	int rn, n1, nn;
	Tablecell *c1, *cn;

	n1 = 0;
	for(c1 = tab->rows[r1].cells; c1; c1 = c1->nextinrow)
		n1++;

	for(rn = r1; rn < tab->nrow; rn++){

		nn = 0;
		for(cn = tab->rows[rn].cells; cn; cn = cn->nextinrow)
			nn++;

		if(n1 != nn)
			return 0;

		c1 = tab->rows[r1].cells;
		cn = tab->rows[rn].cells;
		while(c1 && cn){
			if(c1->align.halign != cn->align.halign)
				return 0;
			if(c1->align.valign != cn->align.valign)
				return 0;
			if(c1->rowspan != cn->rowspan)
				return 0;
			if(c1->colspan != cn->colspan)
				return 0;

			c1 = c1->nextinrow;
			cn = cn->nextinrow;
		}
	}
	return 1;
}

static void
caption(Xlate *x, Item *item, int width)
{
	char *str;
	Xlate xlate;

	xlate = *x;
	xlate.f = &xlate.fmt;
	xlate.width = width;
	xlate.embed = 1;
	xlate.pos = 0;

	fmtstrinit(xlate.f);
	walk(&xlate, item);
	str = fmtstrflush(xlate.f);
	raw(x, str);
	free(str);
}

/* write a table spec, and the cells within it */
static void
table(Xlate *x, Item *item)
{
	Item *ti;
	Table *tab;
	Itable *ita;
	Tablecell *ce;
	int r, idx, hdr;

	ita = (Itable *)item;
	tab = ita->table;

	if(tab->caption && tab->caption_place == ALtop)
		caption(x, tab->caption, x->width);

	/* table layout def */
	lnbrk(x);
	raw(x, ".TS\n");
	if(dimen(x, tab->width) >= x->width*72)	
		raw(x, "box expand ;\n");
	else
		raw(x, "box  ;\n");

// FIXME: need to push font size and font style into table header

	for(r = 0; r < tab->nrow; r++){
		for(ce = tab->rows[r].cells; ce; ce = ce->nextinrow){
			rowdef(x, ce);
			if(ce->nextinrow)
				raw(x, "|");
		}
		if(r >= tab->nrow-1 || allsame(tab, r))
			break;
		raw(x, "\n");
	}
	raw(x, " .\n");

	/* table data */
	for(r = 0; r < tab->nrow; r++){

		idx = 0;
		hdr = 0;
		for(ce = tab->rows[r].cells; ce; ce = ce->nextinrow){

			if(ce->flags & TFisth)
				hdr = 1;
			ti = ce->content;
			cell(x, ti, x->width, idx++, ce->flags&TFnowrap); /* FIXME: should be the width of the cell */
		}
		lnbrk(x);
		if(hdr)
			raw(x, "=\n");
	}
	raw(x, ".TE\n");

	if(tab->caption && tab->caption_place == ALbottom)
		caption(x, tab->caption, x->width);

}
/* write a picture (as a small table) */
static void
image(Xlate *x, Item *item)
{
	Iimage *im;

	if(! aflag)
		return;

	im = (Iimage *)item;
	lnbrk(x);
	raw(x, ".BX ");
	copy(x, im->imsrc);
	raw(x, "\n");
}

/* driver which walks a page hierarchy or part of one */
static void
walk(Xlate *x, Item *items)
{
	Item *il;
	Irule *ir;
	Anchor *a;
	Ifloat *ifl;
	Ispacer *is;
	char buf[32];

	for(il=items; il!=nil; il=il->next){
		switch(il->tag){
		case Itexttag:
				text(x, il);
			break;
		case Iruletag:
			lnbrk(x);
			ir = (Irule *)il;
			snprint(buf, sizeof(buf), "\\l'%dp'\n", dimen(x, ir->wspec));
			raw(x, buf);
			break;
		case Iformfieldtag:
			break;
		case Itabletag:
			table(x, il);
			break;
		case Ifloattag:
			ifl = (Ifloat*)il;
			walk(x, ifl->item);
			break;
		case Iimagetag:
			image(x, il);
			break;
		case Ispacertag:
			is = (Ispacer*)il;
			switch(is->spkind){
			case ISPvline:
				if(x->just == Verbatim){
					raw(x, "\n");
				}
				else{
					/* rather strange logic but seems to be correct */
					if(sol(x))
						raw(x, ".sp 1\n");
					else
						raw(x, "\n.br\n");
				}

				break;
			case ISPhspace:
				raw(x, " ");
				break;
			case ISPgeneral:
				raw(x, " ");
				break;
			case ISPnull:
				raw(x, "\\ ");
				break;
			}
			break;
		default:
			fprint(2, "%s: libhtml error, unknown item tag %d\n", argv0, il->tag);
		}

		if(il->anchorid != 0 && il->anchorid != x->curanchor){
			for(a = x->u->docinfo->anchors; a; a = a->next)
				if(aflag && a->index == il->anchorid){
					lnbrk(x);
					raw(x, ".BX ");
					snprint(buf, sizeof(buf), "url%d", a->index);
					raw(x, buf);
					raw(x, "\n");
					break;
				}
			x->curanchor = il->anchorid;
		}
	}
}

static void
reflist(Xlate *x)
{
	Anchor *a;
	char buf[16], *href;

	if(! aflag)
		return;

	raw(x, ".SH\n");
	raw(x, "URLs\n");
	raw(x, ".LP\n");

	raw(x, ".TS\n");
	raw(x, "allbox expand ;\n");
	raw(x, "l l .\n");

	for(a = x->u->docinfo->anchors; a; a = a->next){
		if(a->href == nil)
			continue;

		href = fullurl(x->u, a->href);
		lnbrk(x);
		snprint(buf, sizeof(buf), "url%d", a->index);
		raw(x, buf);
//		raw(x, "\t");
//		if(a->name)
//			copy(x, a->name);
		raw(x, "\tT{\n");
		copy(x, a->href);
		raw(x, "\nT}\n");
		free(href);
	}
	raw(x, ".TE\n");
}

/* output document headers/trailers, parse http data structure */
void
html2ms(URLwin *u, int width)
{
	char *str;
	Xlate xlate;

	memset(&xlate, 0, sizeof(xlate));
	xlate.f = &xlate.fmt;
	xlate.just = IFwrap;
	xlate.width = width;
	xlate.font = DefFnt;
	xlate.u = u;

	fmtstrinit(xlate.f);

	fmtprint(xlate.f, "\\\" Generated by %s\n", argv0);
	fmtprint(xlate.f, "\\\"\n");

	fmtprint(xlate.f, ".FP lucidasans\n");
	if(u->docinfo){
		if(u->docinfo->doctitle){
			fmtprint(xlate.f, ".TL\n");
			fmtprint(xlate.f, "%S\n", u->docinfo->doctitle);
			fmtprint(xlate.f, ".LP\n");
		}

		if(u->docinfo->src){
			fmtprint(xlate.f, ".FS\n");
			fmtprint(xlate.f, "Generated by %s from %S\n", argv0, u->docinfo->src);
			fmtprint(xlate.f, ".FE\n");
		}
	}

	/*
	 * NB: because we use fmtprint above, we must take care we call walk()
	 * with xlate->pos correctly initialised (0 since we the above ends with \n
	 */
	walk(&xlate, u->items);
	lnbrk(&xlate);

	reflist(&xlate);
	lnbrk(&xlate);

	str = fmtstrflush(xlate.f);
	write(1, str, strlen(str));
	free(str);
}

