#include <u.h>
#include <libc.h>
#include <bio.h>
#include <draw.h>
#include <html.h>
#include <ctype.h>
#include "dat.h"

enum {
	WideTableField = 20,	/* table fields wider than this get wrapped */
	Verbatim = 0,			/* layout preserved, no line wrap, no justification */
};

typedef struct Xlate Xlate;
struct Xlate {
	Fmt fmt;
	Fmt *f;				/* pointer to the above - just makes the syntax cleaner */
	int width;			/* width of this block (may be a table cell) */
	int just;			/* justification mode l/r/c/0 0==Verbatim */
	int pos;			/* output position (column) */
	int curanchor;		/* current anchor ID */
	URLwin *u;			/* urlwin passed in */
};

static void walk(Xlate *x, Item *items);

/* write a single char (no escaping) */
static void
chr(Xlate *x, char c)
{
	if(c == '\n')
		x->pos = 0;
	else
		x->pos++;
	fmtprint(x->f, "%c", c);
}

/* write raw data (no escaping) */
static void
raw(Xlate *x, char *p)
{
	for(;*p; p++)
		chr(x, *p);
}

/*
 * used for prose text which can be re-flowed
 *
 * mutate all whitespace to space characters,
 * drop whitespace sta start of lines,
 * wrap lines on a space char
 */
static void
wrap(Xlate *x, Rune *r)
{
	Rune c;

	if(debug)
		fprint(2, "wrap %3d '%S'\n", x->pos, r);

	for( ;*r; r++){
		c = *r;

		if(isspacerune(c)){
			c = ' ';
			if(x->pos == 0)
				continue;
			if(x->pos > x->width){				/* width from command line */
				fmtprint(x->f, "\n");
				x->pos = 0;
				continue;
			}
		}
		fmtprint(x->f, "%C", c);
		x->pos++;
	}
}

/* used to output verbatim text */
static void
copy(Xlate *x, Rune *r)
{
	Rune c;

	if(debug)
		fprint(2, "copy %3d '%S'\n", x->pos, r);

	for( ;*r; r++){
		c = *r;

	if(c == L'\n'){
			x->pos = 0;
		}else
			x->pos++;
		fmtprint(x->f, "%C", c);
	}
}


/* linebreak (conditional newline) */
static void
lnbrk(Xlate *x)
{
	if(x->pos)
		chr(x, '\n');
}

/* write text blocks */
static void
text(Xlate *x, Item *item)
{
	Itext *it;
	int just, state;

	state = item->state;
	it = (Itext *)item;
	just = state & (IFcjust|IFrjust|IFwrap);

	if(state & IFbrksp){				/* insert a blank line */
		lnbrk(x);
		chr(x, '\n');
	}
	else
	if(state & IFbrk)					/* break current line */
		lnbrk(x);

	if(just == Verbatim)
		copy(x, it->s);
	else
		wrap(x, it->s);
}

static void
caption(Xlate *x, Item *item, int width)
{
	char *str;
	Xlate xlate;

	xlate = *x;
	xlate.f = &xlate.fmt;
	xlate.width = width;
	xlate.pos = 0;

	fmtstrinit(xlate.f);
	walk(&xlate, (Itext *)item);
	str = fmtstrflush(xlate.f);
	raw(x, str);
	free(str);
}

/* write a table cell */
static int
cell(Xlate *x, Item *item, int offset, int width)
{
	int i;
	Xlate xlate;
	char *p, *e, *str;

	xlate = *x;
	xlate.f = &xlate.fmt;
	xlate.width = x->width;		/* should be the width of the field not the width of the doc */
	xlate.pos = 0;

	fmtstrinit(xlate.f);
	walk(&xlate, (Itext *)item);
	str = fmtstrflush(xlate.f);

	p = str;
	if((e = strchr(p, '\n')) == nil){
		raw(x, p);
		for(i = strlen(p); i < width; i++)
			chr(x, ' ');

	}
	else{
		do{
			*e = 0;
			raw(x, p);
			chr(x, '\n');
			for(i = 0; i < offset; i++)
				chr(x, ' ');
			p = e+1;
		}while((e = strchr(p, '\n')) != nil);
	}

	free(str);
	return 1;
}

/* find the width of the text in a cell */
static int
cellwidth(Xlate *x, Item *item)
{
	int width;
	Xlate xlate;
	char *p, *e, *str;

	xlate = *x;
	xlate.f = &xlate.fmt;
	xlate.width = x->width;
	xlate.pos = 0;

	fmtstrinit(xlate.f);
	walk(&xlate, (Itext *)item);
	str = fmtstrflush(xlate.f);

	p = str;
	if((e = strchr(p, '\n')) == nil){
		width = strlen(p);
	}
	else{
		width = 0;
		do{
			*e = 0;
			if(e-p > width)
				width = e-p;
			p = e+1;
		}while((e = strchr(p, '\n')) != nil);
	}
	free(str);
	return width;
}

/* write a table spec, and the cells within it */
static void
table(Xlate *x, Item *item)
{
	Table *tab;
	Itable *ita;
	Tablerow *ro;
	Tablecell *ce;
	int r, c, n, offset, *widths;

	ita = (Itable *)item;
	tab = ita->table;

	widths = emalloc(sizeof(*widths)*tab->ncol);

	if(tab->caption && tab->caption_place == ALtop)
		caption(x, tab->caption, x->width);

	/* get column widths, ignore the first row - it may be un-reprisentitive */
	memset(widths, 0, sizeof(widths));
	for(r = 1; r < tab->nrow; r++)
		for(c = 0, ce = tab->rows[r].cells; ce != nil; ce = ce->nextinrow, c++){
			n = cellwidth(x, ce->content);
			if(n > widths[c])
				widths[c] = n;
		}

	/* now add first row if we got nothing else */
	for(c = 0, ce = tab->rows[0].cells; ce != nil; ce = ce->nextinrow, c++){
		n = cellwidth(x, ce->content);
		if(widths[c] == 0)
			widths[c] = n;
	}

	lnbrk(x);
	for(r = 0; r < tab->nrow; r++){
		ro = &tab->rows[r];

		offset = 0;
		for(c = 0, ce = ro->cells; ce != nil; ce = ce->nextinrow, c++){
			cell(x, ce->content, offset, (ce->nextinrow)? widths[c]+2: 0);
			offset += widths[c]+2;
		}
		lnbrk(x);
	}

	free(widths);

	if(tab->caption && tab->caption_place == ALbottom)
		caption(x, tab->caption,  x->width);
}

static void
image(Xlate *x, Item *item)
{
	Iimage *im;

	if(! aflag)
		return;

	im = (Iimage *)item;
	raw(x, " [image ");
	copy(x, im->imsrc);
	if(im->altrep){
		raw(x, " - ");
		copy(x, im->altrep);
	}
	raw(x, "]\n");
}


/* driver which walks a page hierarchy or part of one */
static void
walk(Xlate *x, Item *items)
{
	int i;
	Item *il;
	Anchor *a;
	Ifloat *ifl;
	Ispacer *is;
	char *href;

	for(il = items; il != nil; il = il->next){
		switch(il->tag){
		case Itexttag:
			text(x, il);
			break;
		case Iruletag:
			lnbrk(x);
			for(i = 0; i < x->width; i++)
				chr(x, '_');
			chr(x, '\n');
			break;
		case Iformfieldtag:
			break;
		case Itabletag:
			table(x, il);
			break;
		case Ifloattag:
			ifl = (Ifloat*)il;
			walk(x, ifl->item);
			break;
		case Iimagetag:
			image(x, il);
			break;
		case Ispacertag:
			is = (Ispacer*)il;
			switch(is->spkind){
			case ISPvline:
				chr(x, '\n');
				break;
			case ISPhspace:
				chr(x, ' ');
				break;
			case ISPgeneral:
				chr(x, ' ');
				break;
			case ISPnull:
				chr(x, ' ');
				break;
			}
			break;
		default:
			fprint(2, "%s: libhtml error, unknown item tag %d\n", argv0, il->tag);
		}

		if(il->anchorid != 0 && il->anchorid != x->curanchor){
			for(a = x->u->docinfo->anchors; a; a = a->next)
				if(aflag && a->index == il->anchorid){
					href = fullurl(x->u, a->href);
					lnbrk(x);
					raw(x, " [");
					raw(x, href);
					raw(x, "]\n");
					free(href);
					break;
				}
			x->curanchor = il->anchorid;
		}
	}
}

void
htmlfmt(URLwin *u, int width)
{
	char *str;
	Xlate xlate;

	memset(&xlate, 0, sizeof(xlate));
	xlate.f = &xlate.fmt;
	xlate.just = IFwrap;
	xlate.width = width;
	xlate.u = u;

	fmtstrinit(xlate.f);
	walk(&xlate, u->items);
	lnbrk(&xlate);
	str = fmtstrflush(xlate.f);
	write(1, str, strlen(str));
	free(str);
}

